#define DOLPHOT_C
#include <dolphot.h>
char camera[81]="";

int ferr(FILE *f,char *str) {
   printf("%s\n",str);
   if (f!=NULL) fclose(f);
   exit(1);
}

void merr(void) {
   printf("Memory allocation error\n");
   exit(1);
}

chiptype allocchip(int X,int Y) {
   chiptype tmp;
   int i;

   if (X<1) X=1;
   if (Y<1) Y=1;
   tmp=(float**)calloc(Y,sizeof(float*));
   if (!tmp) merr();
   tmp[0]=(float*)calloc((long)Y*(long)X,sizeof(float));
   if (!tmp[0]) merr();
   for (i=1;i<Y;i++) tmp[i]=tmp[i-1]+X;
   return tmp;
}

void freechip(chiptype ptr,int X,int Y) {
   free(ptr[0]);
   free(ptr);
   return;
}

imgtype allocimg(int X,int Y,int Z) {
   imgtype tmp;
   int i;

   if (Z<1) Z=1;
   tmp=(float***)calloc(Z,sizeof(float**));
   if (!tmp) merr();
   for (i=0;i<Z;i++) tmp[i]=allocchip(X,Y);
   return tmp;
}

void freeimg(imgtype tmp,int X,int Y,int Z) {
   int i;

   if (Z<1) Z=1;
   for (i=0;i<Z;i++) freechip(tmp[i],X,Y);
   free(tmp);
}

// endian-testing code
short int ENDIAN_TEST_WORD=1;
char *IS_LITTLE_ENDIAN=(char*)(&ENDIAN_TEST_WORD);

size_t ffread(void *ptr,size_t size,size_t N,FILE *f) {
   size_t rv;
   rv=fread(ptr,size,N,f);
   if (*IS_LITTLE_ENDIAN==1) {
      long tot,i;
      char *cp,x;
      cp=(char*)ptr;
      tot=N*size;
      for (i=0;i<tot;i+=4) {
	 x=cp[i]; cp[i]=cp[i+3]; cp[i+3]=x;
	 x=cp[i+1]; cp[i+1]=cp[i+2]; cp[i+2]=x;
      }
   }
   return rv;
}

size_t ffwrite(const void *ptr,size_t size,size_t N,FILE *f) {
   if (*IS_LITTLE_ENDIAN==0) {
      return fwrite(ptr,size,N,f);
   }
   char *cp,x;
   long tot,i;
   size_t rv;
   tot=N*size;
   cp=(char*)calloc(tot,1);
   if (!cp) {
      printf("Memory error in ffwrite\n");
      exit(0);
   }
   memcpy(cp,ptr,tot);
   for (i=0;i<tot;i+=4) {
      x=cp[i]; cp[i]=cp[i+3]; cp[i+3]=x;
      x=cp[i+1]; cp[i+1]=cp[i+2]; cp[i+2]=x;
   }
   rv=fwrite(cp,size,N,f);
   free(cp);
   return rv;
}

size_t sfread(void *ptr,size_t size,size_t N,FILE *f) {
   size_t rv;
   rv=fread(ptr,size,N,f);
   if (*IS_LITTLE_ENDIAN==1) {
      long tot,i;
      char *cp,x;
      cp=(char*)ptr;
      tot=N*size;
      for (i=0;i<tot;i+=2) {
	 x=cp[i]; cp[i]=cp[i+1]; cp[i+1]=x;
      }
   }
   return rv;
}

size_t sfwrite(const void *ptr,size_t size,size_t N,FILE *f) {
   if (*IS_LITTLE_ENDIAN==0) {
      return fwrite(ptr,size,N,f);
   }
   char *cp,x;
   long tot,i;
   size_t rv;
   tot=N*size;
   cp=(char*)calloc(tot,1);
   if (!cp) {
      printf("Memory error in sfwrite\n");
      exit(0);
   }
   memcpy(cp,ptr,tot);
   for (i=0;i<tot;i+=2) {
      x=cp[i]; cp[i]=cp[i+1]; cp[i+1]=x;
   }
   rv=fwrite(cp,size,N,f);
   free(cp);
   return rv;
}

int parseparam(char*str,int(*func)(char*,char*)) {
   char var[256],val[256],*ptr;

   for (ptr=str;*ptr==32 || *ptr==9;ptr++);
   if (*ptr=='#') return 0;
   strcpy(var,ptr);
   if ((ptr=strchr(var,'='))==NULL) {
      printf("Cannot parse parameter line:\n%s\n",str);
      return -2;
   }
   *ptr=0;
   ptr++;
   for (;*ptr==32 || *ptr==9;ptr++);
   strcpy(val,ptr);
   if ((ptr=strchr(val,'#'))!=NULL) *ptr=0;
   for (ptr=var+strlen(var);ptr>var && (*(ptr-1)==32 || *(ptr-1)==9);ptr--);
   *ptr=0;
   for (ptr=val+strlen(val);ptr>val && (*(ptr-1)==32 || *(ptr-1)==9 || *(ptr-1)==10 || *(ptr-1)==13);ptr--);
   *ptr=0;
   if ((*func)(var,val)) return 0;
   printf("Cannot match parameter name:\n%s\n",var);
   return -1;
}

void paramfile(char*fn,int(*func)(char*,char*)) {
   char str[256];
   FILE *f;
   int any=0;

   sprintf(str,"%s/param/%s",BASEDIR,fn);
   if ((f=fopen(str,"r"))!=NULL) {
      any=1;
      while (fgets(str,256,f)) parseparam(str,func);
      fclose(f);
   }
   if (camera[0]) {
      sprintf(str,"%s/%s/%s",BASEDIR,camera,fn);
      if ((f=fopen(str,"r"))!=NULL) {
	 any=1;
	 while (fgets(str,256,f)) parseparam(str,func);
	 fclose(f);
      }
   }
   if ((f=fopen(fn,"r"))!=NULL) {
      any=1;
      while (fgets(str,256,f)) parseparam(str,func);
      fclose(f);
   }
   if (!any) printf("No parameter file \"%s\" found; using defaults\n",fn);
   return;
}

void paramfile1(char*fn,int(*func)(char*,char*)) {
   char str[256];
   FILE *f;

   if ((f=fopen(fn,"r"))!=NULL) {
      while (fgets(str,256,f)) parseparam(str,func);
      fclose(f);
   }
   else printf("No parameter file \"%s\" found; using defaults\n",fn);
   return;
}
