#define FITS_C
#include <fits.h>

typedef int int32;
typedef short int16;
typedef char int8;

int fitsparam(char*var,char*val) {
   double x;
   char *ptr;
   if (strlen(val)<=8) {
      if (!strcasecmp(var,"gain_kw")) {sprintf(fitsstr[0],"%-8s=",val); return 1;}
      if (!strcasecmp(var,"exptime_kw")) {sprintf(fitsstr[1],"%-8s=",val); return 1;}
      if (!strcasecmp(var,"rdnoise_kw")) {sprintf(fitsstr[2],"%-8s=",val); return 1;}
      if (!strcasecmp(var,"minval_kw")) {sprintf(fitsstr[3],"%-8s=",val); return 1;}
      if (!strcasecmp(var,"maxval_kw")) {sprintf(fitsstr[4],"%-8s=",val); return 1;}
      if (!strcasecmp(var,"mjd_kw")) {sprintf(fitsstr[5],"%-8s=",val); return 1;}
      if (!strcasecmp(var,"airmass_kw")) {sprintf(fitsstr[6],"%-8s=",val); return 1;}
      if (!strcasecmp(var,"exptime0_kw")) {sprintf(fitsstr[7],"%-8s=",val); return 1;}
   }
   x=strtod(val,&ptr);
   if (!*ptr) {
      if (!strcasecmp(var,"gain_val")) {fitsval[0]=x; return 1;}
      if (!strcasecmp(var,"exptime_val")) {fitsval[1]=x; return 1;}
      if (!strcasecmp(var,"rdnoise_val")) {fitsval[2]=x; return 1;}
      if (!strcasecmp(var,"minval_val")) {fitsval[3]=(double)((float)x); return 1;}
      if (!strcasecmp(var,"maxval_val")) {fitsval[4]=(double)((float)x); return 1;}
      if (!strcasecmp(var,"mjd_val")) {fitsval[5]=x; return 1;}
      if (!strcasecmp(var,"airmass_val")) {fitsval[6]=x; return 1;}
      if (!strcasecmp(var,"exptime0_val")) {fitsval[7]=x; return 1;}
   }
   return 0;
}

void addcard(imtype*img,char str[81]) {
   int i;

   if (!strncmp(str,"COMMENT  ",9)) i=img->Ncards;
   else for (i=0;i<img->Ncards && strncmp(str,img->cards[i],9);i++);
   if (i<img->Ncards) strcpy(img->cards[i],str);
   else {
      if (i>=img->Nmax) {
	 cardtype *tmp;
	 int newNmax = i+14;
	 tmp=(cardtype*)calloc(sizeof(cardtype),newNmax);
	 if (!tmp) merr();
	 if (img->Ncards) {
	    memcpy(tmp,(img->cards),sizeof(cardtype)*(img->Ncards));
	 }
	 if (img->Nmax) {
	    free(img->cards);
	 }
	 img->cards=tmp;
	 img->Nmax = newNmax;
      }
      strcpy(img->cards[i],str);
      img->Ncards=i+1;
   }
   return;
}

void skipimage(FILE *f,imtype*img) {
   int skip,z,y;
   char ch[2880],*buf;

   buf=(char*)calloc(img->X,abs(img->bits)/8);
   if (!buf) merr();
   for (z=0;z<img->Z;z++) for (y=0;y<img->Y;y++) fread(buf,abs(img->bits)/8,img->X,f);
   skip=abs(img->bits)/8*img->X*img->Y*img->Z;
   skip=((skip+2879)/2880)*2880-skip;
   fread(ch,1,skip,f);
   free(buf);
}

int readcards(FILE *f,ftype*fits,imtype*img) {
   int MODE=0,line=0,tmp,EXT=0;
   cardtype str="";
   int ext_set=0;

   img->bits=-32;
   img->bzero=0.;
   img->bscale=1.;
   img->pcount=0;
   strcpy(img->xtension,"IMAGE");
   img->X=img->Y=img->Z=1;
   img->Ncards=0;
   img->cards=0;
   img->Nmax=0;
   if (fits) fits->Next=-1;
   while (!feof(f) && (!MODE || line%36)) {
      //fgets(str,81,f);
      fread(str,1,80,f); str[80]=0;
      line++;
      if (!MODE) {
	 if (!strncmp(str,"END     ",8)) MODE=1;
	 else if (!strncmp(str,"SIMPLE  =",9)) {
	    if (str[29]!='T') ferr(f,"Illegal format type -- must be simple");
	    if (!fits) ferr(f,"SIMPLE should not be in extensions");
	 }
	 else if (!strncmp(str,"NAXIS   =",9)) {
	    if ((tmp=atoi(str+9))==2) img->Z=1;
	    else if (tmp==1) img->Y=img->Z=1;
	    else if (tmp==0) img->X=img->Y=img->Z=0;
	    else if (tmp!=3) ferr(f,"Illegal number of axes");
	 }
	 else if (!strncmp(str,"NAXIS",5)) {
	    tmp=atoi(str+9);
	    if (str[5]=='1') img->X=tmp;
	    else if (str[5]=='2') img->Y=tmp;
	    else if (str[5]=='3') img->Z=tmp;
	    else ferr(f,"Illegal Axis Number");
	 }
	 else if (!strncmp(str,"BITPIX  =",9)) img->bits=atoi(str+9);
	 else if (!strncmp(str,"BSCALE  =",9)) img->bscale=atof(str+9);
	 else if (!strncmp(str,"BZERO   =",9)) img->bzero=atof(str+9);
	 else if (!strncmp(str,"PCOUNT  =",9)) img->pcount=atoi(str+9);
	 else if (!strncmp(str,"EXTEND  =",9)) {
	    if (fits==NULL) ferr(f,"Extension format error");
	    else if (str[29]=='T') EXT=1;
	 }
	 else if (!strncmp(str,"NEXTEND =",9)) {
	    if (EXT) {
	       if (fits==NULL) ferr(f,"Extension format error");
	       fits->Next=atoi(str+9);
	    }
	 }
	 else if (!strncmp(str,"XTENSION=",9)) {
	    if (fits!=NULL) ferr(f,"Extension format error");
	    for (tmp=9;str[tmp] && str[tmp]!='\'';tmp++);
	    if (!str[tmp]) ferr(f,"Cannot read extension type");
	    tmp++;
	    while (str[tmp]==32) tmp++;
	    strcpy(img->xtension,str+tmp);
	    for (tmp=0;img->xtension[tmp] && img->xtension[tmp]!='\'';tmp++);
	    if (!img->xtension[tmp]) ferr(f,"Cannot read extension type");
	    tmp--;
	    while (tmp>=0 && img->xtension[tmp]==32) tmp--;
	    img->xtension[tmp+1]=0;
	    ext_set=1;
	 }
	 else addcard(img,str);
      }
   }
   // if NEXTEND not set
   if (fits && fits->Next<0) {
      fits->Next=0;
      if (EXT) {
	 long fpos;
	 imtype img;
	 fpos=ftell(f);
	 skipimage(f,&(fits->img));
	 while (readcards(f,NULL,&img)) {
	    skipimage(f,&img);
	    fits->Next++;
	 }
	 fseek(f,fpos,SEEK_SET);
      }
   }
   return ext_set;
}

FILE *readfitsh(char*fn,ftype*fits,int ver) {
   FILE *f;
   static int first=1;

   if ((f=fopen(fn,"rb"))==NULL) {
      printf("Error reading \"%s\"\n",fn);
      exit(1);
   }
   if (first) {
      paramfile("fits.param",&fitsparam);
      first=0;
   }
   readcards(f,fits,&(fits->img));
   if (ver) {
      if (fits->img.Z>1) printf("Reading FITS file %s: %dx%dx%d\n",fn,fits->img.X,fits->img.Y,fits->img.Z);
      else if (fits->img.Y>1) printf("Reading FITS file %s: %dx%d\n",fn,fits->img.X,fits->img.Y);
      else if (fits->img.X>0) printf("Reading FITS file %s: %d\n",fn,fits->img.X);
      else printf("Reading FITS file %s\n",fn);
   }
   return f;
}

void readexth(FILE *f,imtype*img,int ver) {
   readcards(f,NULL,img);
   if (ver) {
      if (img->Z>1) printf("Reading %s extension: %dx%dx%d\n",img->xtension,img->X,img->Y,img->Z);
      else if (img->Y>1) printf("Reading %s extension: %dx%d\n",img->xtension,img->X,img->Y);
      else if (img->X>0) printf("Reading %s extension: %d\n",img->xtension,img->X);
      else printf("Reading %s extension\n",img->xtension);
   }
   return;
}

void readchip(FILE *f,chiptype chip,imtype*img) {
   int i,j;
   int32 l;
   int16 s;
   int8 c;
   float *fp,*fpmax;

   if (sizeof(float)!=4 || sizeof(int32)!=4 || sizeof(int16)!=2 || sizeof(int8)!=1) {
      printf("Incorrect data sizes in fits_lib.c\n");
      exit(-1);
   }
   if (img->bits==-32) {
      ffread(chip[0],(img->X)*(img->Y),4,f);
      if (img->bscale==1. && img->bzero==0.) return;
      fp=chip[0]; fpmax=fp+(img->X)*(img->Y);
      while (fp<fpmax) (*fp)=(*fp)*img->bscale+img->bzero;
      return;
   }
   for (i=0;i<img->Y;i++) {
      if (img->bits==32) {
	 for (j=0;j<img->X;j++) {
	    ffread(&l,1,4,f);
	    chip[i][j]=l*img->bscale+img->bzero;
	 }
      }
      else if (img->bits==16) {
	 for (j=0;j<img->X;j++) {
	    sfread(&s,1,2,f);
	    chip[i][j]=s*img->bscale+img->bzero;
	 }
      }
      else if (img->bits==8) {
	 for (j=0;j<img->X;j++) {
	    fread(&c,1,1,f);
	    chip[i][j]=c*img->bscale+img->bzero;
	 }
      }
   }
   return;
}

void readimage(FILE *f,imtype*img) {
   int skip,z;
   char ch[2880];

   img->data=allocimg(img->X,img->Y,img->Z);
   for (z=0;z<img->Z;z++) readchip(f,img->data[z],img);
   skip=abs(img->bits)/8*img->X*img->Y*img->Z+img->pcount;
   skip=((skip+2879)/2880)*2880-skip;
   fread(ch,1,skip,f);
}

void readbody(FILE *f,ftype*fits,int ver) {
   int i;

   readimage(f,&(fits->img));
   if (fits->Next>0) {
      fits->ext=(imtype*)calloc(sizeof(imtype),fits->Next);
      if (!fits->ext) merr();
      for (i=0;i<fits->Next;i++) {
	 readexth(f,fits->ext+i,ver);
	 readimage(f,&(fits->ext[i]));
      }
   }
   fclose(f);
   return;
}

void readfits(char*fn,ftype*fits,int ver) {
   readbody(readfitsh(fn,fits,ver),fits,ver);
   return;
}

void readfitsinfo(char*fn,ftype*fits,int ver) {
   FILE *f;
   int i;

   f=readfitsh(fn,fits,ver);
   skipimage(f,&(fits->img));
   if (fits->Next>0) {
      fits->ext=(imtype*)calloc(sizeof(imtype),fits->Next);
      if (!fits->ext) merr();
      for (i=0;i<fits->Next;i++) {
	 readexth(f,fits->ext+i,ver);
	 skipimage(f,&(fits->ext[i]));
      }
   }
   fclose(f);
   return;
}

void parsecards(imtype*img,double*GAIN,double*RN,double*EXP,float*DMIN,float*DMAX,double*EPOCH,double*AIR,double*EXP0,int ver,int usedef) {
   int i;

   if (usedef) {
      if (GAIN!=NULL) *GAIN=fitsval[0];
      if (EXP!=NULL) *EXP=fitsval[1];
      if (RN!=NULL) *RN=fitsval[2];
      if (DMIN!=NULL) *DMIN=fitsval[3];
      if (DMAX!=NULL) *DMAX=fitsval[4];
      if (EPOCH!=NULL) *EPOCH=fitsval[5];
      if (AIR!=NULL) *AIR=fitsval[6];
      if (EXP0!=NULL) *EXP0=fitsval[7];
   }
   for (i=0;i<img->Ncards;i++) {
      if (GAIN!=NULL && !strncmp(img->cards[i],fitsstr[0],9)) *GAIN=atof(img->cards[i]+9);
      else if (EXP!=NULL && !strncmp(img->cards[i],fitsstr[1],9)) *EXP=atof(img->cards[i]+9);
      else if (RN!=NULL && !strncmp(img->cards[i],fitsstr[2],9)) *RN=atof(img->cards[i]+9);
      else if (DMIN!=NULL && !strncmp(img->cards[i],fitsstr[3],9)) *DMIN=atof(img->cards[i]+9);
      else if (DMAX!=NULL && !strncmp(img->cards[i],fitsstr[4],9)) *DMAX=atof(img->cards[i]+9);
      else if (EPOCH!=NULL && !strncmp(img->cards[i],fitsstr[5],9)) *EPOCH=atof(img->cards[i]+9);
      else if (AIR!=NULL && !strncmp(img->cards[i],fitsstr[6],9)) *AIR=atof(img->cards[i]+9);
      else if (EXP0!=NULL && !strncmp(img->cards[i],fitsstr[7],9)) *EXP0=atof(img->cards[i]+9);
   }
   if (ver) {
      printf(" ");
      if (GAIN!=NULL) printf(" GAIN=%4.2f",*GAIN);
      if (EXP!=NULL) printf(" EXP=%ds",(int)*EXP);
      if (RN!=NULL) printf(" NOISE=%4.2f",*RN);
      if (DMIN!=NULL) printf(" BAD=%4.2f",*DMIN);
      if (DMAX!=NULL) printf(" SAT=%4.2f",*DMAX);
      printf("\n");
   }
}

char*getcard(int Ncards,cardtype*cards,char*str) {
   char sstr[10];
   static cardtype card;
   int i;

   if (strlen(str)==9 && str[8]=='=') strcpy(sstr,str);
   else sprintf(sstr,"%-8s=",str);
   for (i=0;i<Ncards;i++) if (!strncmp(cards[i],sstr,9)) {
      strcpy(card,cards[i]);
      return card;
   }
   strcpy(card,"                                                                                ");
   return card;
}

char*getcardval(imtype*img,char*val,int err) {
   static char str[2][81];
   static int nstr=0;
   char card[81];
   int i;

   nstr=1-nstr;
   str[nstr][0]=0;
   strcpy(card,getcard(img->Ncards,img->cards,val));
   if (!strcmp(card,"                                                                                ")) {
      if (err) printf("Error: FITS card \"%s\" not found\n",val);
      return str[nstr];
   }
   if (strlen(card)<9) return str[nstr];
   i=9;
   while (card[i]==' ') i++;
   if (card[i]=='\'') {
      i++;
      strcpy(str[nstr],card+i);
      i=0;
      while (str[nstr][i] && str[nstr][i]!='\'') i++;
      while (i>0 && str[nstr][i-1]==32) i--;
      str[nstr][i]=0;
      return str[nstr];
   }
   strcpy(str[nstr],card+i);
   i=0;
   while (str[nstr][i] && str[nstr][i]!='/') i++;
   while (i>0 && str[nstr][i-1]==32) i--;
   str[nstr][i]=0;
   return str[nstr];
}

char* fitsnum(double x) {
   static char str[401];
   char *ptr;
   sprintf(str,"%20f",x);
   if (strlen(str)<=20) return str;
   sprintf(str,"%20.13e",x);
   if (strlen(str)>20) sprintf(str,"%20.12e",x);
   if (strlen(str)>20) sprintf(str,"%20.10e",x);
   ptr=strchr(str,'e');
   if (ptr!=NULL) *ptr='E';
   return str;
}

void insertcards(imtype*img,double GAIN,double RN,double EXP,float DMIN,float DMAX,double EPOCH,double AIR,double EXP0) {
   cardtype str;

   if (GAIN>-0.9e30) {
      sprintf(str,"%-9s %20s / Gain (electrons/ADU)                           ",fitsstr[0],fitsnum(GAIN));
      addcard(img,str);
   }
   if (EXP>-0.9e30) {
      sprintf(str,"%-9s %20s / Exposure time (s)                              ",fitsstr[1],fitsnum(EXP));
      addcard(img,str);
   }
   if (RN>-0.9e30) {
      sprintf(str,"%-9s %20s / Read noise (electrons)                         ",fitsstr[2],fitsnum(RN));
      addcard(img,str);
   }
   if (DMIN>-0.9e30) {
      sprintf(str,"%-9s %20s / Low bad data value (DN)                        ",fitsstr[3],fitsnum(DMIN));
      addcard(img,str);
   }
   if (DMAX>-0.9e30) {
      sprintf(str,"%-9s %20s / High bad data value (DN)                       ",fitsstr[4],fitsnum(DMAX));
      addcard(img,str);
   }
   if (EPOCH>-0.9e30) {
      sprintf(str,"%-9s %20s / Epoch (MJD)                                    ",fitsstr[5],fitsnum(EPOCH));
      addcard(img,str);
   }
   if (AIR>-0.9e30) {
      sprintf(str,"%-9s %20s / Air mass                                       ",fitsstr[6],fitsnum(AIR));
      addcard(img,str);
   }
   if (EXP0>-0.9e30) {
      sprintf(str,"%-9s %20s / Single image exposure time (s)                 ",fitsstr[7],fitsnum(EXP0));
      addcard(img,str);
   }
}

void writecards(FILE *f,ftype*fits,imtype*img) {
   char str[81];
   int line=1,i;

   if (fits) {
      fputs("SIMPLE  =                    T / Standard FITS format                           ",f); line++;
   }
   else {
      sprintf(str,"XTENSION= '%-18s' / Extension type                                 ",img->xtension); fputs(str,f); line++;
   }
   sprintf(str,"BITPIX  = %20d / Data type                                      ",img->bits); fputs(str,f); line++;
   if (img->Z>1) {
      fputs("NAXIS   =                    3 / Number of axes                                 ",f); line++;
      sprintf(str,"NAXIS1  = %20d /                                                ",img->X); fputs(str,f); line++;
      sprintf(str,"NAXIS2  = %20d /                                                ",img->Y); fputs(str,f); line++;
      sprintf(str,"NAXIS3  = %20d /                                                ",img->Z); fputs(str,f); line++;
   }
   else if (img->Y>1) {
      fputs("NAXIS   =                    2 / Number of axes                                 ",f); line++;
      sprintf(str,"NAXIS1  = %20d /                                                ",img->X); fputs(str,f); line++;
      sprintf(str,"NAXIS2  = %20d /                                                ",img->Y); fputs(str,f); line++;
   }
   else if (img->X>0) {
      fputs("NAXIS   =                    1 / Number of axes                                 ",f); line++;
      sprintf(str,"NAXIS1  = %20d /                                                ",img->X); fputs(str,f); line++;
   }
   else {
      fputs("NAXIS   =                    0 / Number of axes                                 ",f); line++;
   }
   if (fits) {
      if (fits->Next<=0) {fputs("EXTEND  =                    F / There are no standard extensions               ",f); line++;}
      else {
	 fputs("EXTEND  =                    T / There may be standard extensions               ",f); line++;
	 sprintf(str,"NEXTEND =       %14d /                                                ",fits->Next); fputs(str,f); line++;
      }
   }
   sprintf(str,"BSCALE  = %20s / Image scale                                    ",fitsnum(img->bscale)); fputs(str,f); line++;
   sprintf(str,"BZERO   = %20s / Image zero                                     ",fitsnum(img->bzero)); fputs(str,f); line++;
   for (i=0;i<img->Ncards;i++) {
      sprintf(str,"%-80s",img->cards[i]);
      fputs(str,f);
      line++;
   }
   for (;line%36;line++) fputs("                                                                                ",f);
   fputs("END                                                                             ",f);
}

FILE *writefitsh(char*fn,ftype*fits,int ver) {
   FILE *f;

   if ((f=fopen(fn,"wb"))==NULL) {
      printf("Error writing to \"%s\"\n",fn);
      exit(1);
   }
   writecards(f,fits,&(fits->img));
   if (ver) {
      if (fits->img.Z>1) printf("Writing FITS file %s: %dx%dx%d\n",fn,fits->img.X,fits->img.Y,fits->img.Z);
      else if (fits->img.Y>1) printf("Writing FITS file %s: %dx%d\n",fn,fits->img.X,fits->img.Y);
      else if (fits->img.X>0) printf("Writing FITS file %s: %d\n",fn,fits->img.X);
      else printf("Writing FITS file %s\n",fn);
   }
   return f;
}

void writeexth(FILE *f,imtype*img,int ver) {
   writecards(f,NULL,img);
   if (ver) {
      if (img->Z>1) printf("Writing %s extension: %dx%dx%d\n",img->xtension,img->X,img->Y,img->Z);
      else if (img->Y>1) printf("Writing %s extension: %dx%d\n",img->xtension,img->X,img->Y);
      else if (img->X>0) printf("Writing %s extension: %d\n",img->xtension,img->X);
      else printf("Writing %s extension\n",img->xtension);
   }
   return;
}

void writechip(FILE *f,chiptype chip,imtype*img) {
   int i,j;
   int32 l;
   int16 s;
   int8 c;
   float fl;

   if (sizeof(float)!=4 || sizeof(int32)!=4 || sizeof(int16)!=2 || sizeof(int8)!=1) {
      printf("Incorrect data sizes in fits_lib.c\n");
      exit(-1);
   }
   if (img->bits==-32 && img->bscale==1. && img->bzero==0.) {
      ffwrite(chip[0],(img->X)*(img->Y),4,f);
      return;
   }
   for (i=0;i<img->Y;i++) {
      if (img->bits==-32) {
	 for (j=0;j<img->X;j++) {
	    fl=(chip[i][j]-img->bzero)/img->bscale;
	    ffwrite(&fl,1,4,f);
	 }
      }
      else if (img->bits==32) {
	 for (j=0;j<img->X;j++) {
	    l=(int32)((chip[i][j]-img->bzero)/img->bscale+0.5);
	    ffwrite(&l,1,4,f);
	 }
      }
      else if (img->bits==16) {
	 for (j=0;j<img->X;j++) {
	    s=(int16)((chip[i][j]-img->bzero)/img->bscale+0.5);
	    sfwrite(&s,1,2,f);
	 }
      }
      else if (img->bits==8) {
	 for (j=0;j<img->X;j++) {
	    c=(int8)((chip[i][j]-img->bzero)/img->bscale+0.5);
	    fwrite(&c,1,1,f);
	 }
      }
   }
   fflush(f);
}

void writeimage(FILE *f,imtype*img) {
   int skip,z;
   char ch[2880];

   for (z=0;z<img->Z;z++) writechip(f,img->data[z],img);
   skip=abs(img->bits)/8*img->X*img->Y*img->Z;
   skip=((skip+2879)/2880)*2880-skip;
   memset(ch,0,skip);
   fwrite(ch,1,skip,f);
   fflush(f);
}

void writefits(char*fn,ftype*fits,int ver) {
   FILE *f;
   int i;

   f=writefitsh(fn,fits,ver);
   writeimage(f,&(fits->img));
   for (i=0;i<fits->Next;i++) {
      writeexth(f,fits->ext+i,ver);
      writeimage(f,fits->ext+i);
   }
   fclose(f);
   return;
}

void freeim(imtype*img) {
   if (img->Nmax) free(img->cards);
   img->Nmax=0;
   freeimg(img->data,img->X,img->Y,img->Z);
}

void imcopy(imtype*oldimg,imtype*newimg) {
   int y,z;
   memcpy(newimg,oldimg,sizeof(imtype));
   if (newimg->Nmax) {
      newimg->cards=(cardtype*)calloc(sizeof(cardtype),newimg->Nmax);
      memcpy(newimg->cards,oldimg->cards,sizeof(cardtype)*(newimg->Nmax));
   }
   newimg->data=allocimg(newimg->X,newimg->Y,newimg->Z);
   for (z=0;z<newimg->Z;z++) for (y=0;y<newimg->Y;y++) memcpy(newimg->data[z][y],oldimg->data[z][y],sizeof(float)*(newimg->X));
}

void fitscopy(ftype*oldfits,ftype*newfits) {
   int i;

   newfits->Next=oldfits->Next;
   imcopy(&(oldfits->img),&(newfits->img));
   if (newfits->Next) {
      newfits->ext=(imtype*)calloc(sizeof(imtype),newfits->Next);
      if (!newfits->ext) merr();
      for (i=0;i<newfits->Next;i++) imcopy(oldfits->ext+i,newfits->ext+i);
   }
}

void freefits(ftype*fits) {
   int i;
   freeim(&(fits->img));
   for (i=0;i<fits->Next;i++) freeim(fits->ext+i);
   if (fits->Next) free(fits->ext);
}

void freeinfo(ftype*fits) {
   int i;
   if (fits->img.Nmax) free(fits->img.cards);
   fits->img.Nmax=0;
   for (i=0;i<fits->Next;i++) if (fits->ext[i].Nmax) free(fits->ext[i].cards);
   if (fits->Next) free(fits->ext);
}

int isimage(imtype*img) {
   if (img->Z>1 || img->Y>1) return 1;
   if (strcasecmp(img->xtension,"image")) return 1;
   return 0;
}
